#include <elf.h>
#include <linux/ctype.h>
#include <linux/bitfield.h>

#ifdef CONFIG_CMD_SSBOOT_FS_TYPE
#define SSBOOT_FS_TYPE	CONFIG_CMD_SSBOOT_FS_TYPE
#else
#define SSBOOT_FS_TYPE	FS_TYPE_ANY
#endif

#ifdef CONFIG_CMD_SSBOOT_CKSUM

#define elf_cksum_algorithm(cksum_phdr)		((cksum_phdr)->p_offset)
#define elf_ehdr_cksum(cksum_phdr)		((cksum_phdr)->p_vaddr)
#define elf_phdr_cksum(cksum_phdr)		((cksum_phdr)->p_paddr)
#define elf_data_cksum(cksum_phdr)		((cksum_phdr)->p_filesz)

#define CKSUM_TYPE_CRC32	1

enum {
	SSBOOT_WITH_CRC32,
	SSBOOT_WITHOUT_CKSUM,
};
#endif

static u16 LZ4_readLE16(const void *src) { return le16_to_cpu(*(u16 *)src); }
static void LZ4_copy4(void *dst, const void *src) { *(u32 *)dst = *(u32 *)src; }
static void LZ4_copy8(void *dst, const void *src) { *(u64 *)dst = *(u64 *)src; }

typedef uint8_t  BYTE;
typedef uint16_t U16;
typedef uint32_t U32;
typedef int32_t  S32;
typedef uint64_t U64;

#define FORCE_INLINE static inline __attribute__((always_inline))

#include "../lib/lz4.c"

#define LZ4F_BLOCKUNCOMPRESSED_FLAG 0x80000000U

struct ss_buf {
	void *data;
	u64 start, size;
};
#define buf_len(buf) ((buf)->size - (buf)->start)
#define buf_cur(buf) ((buf)->data + (buf)->start)

struct ss_elf {
	Elf_Ehdr ehdr;
	Elf_Phdr *phdrs;
};

struct ss_ctx {
	int		mode;
	struct ss_elf	elf;
	struct ss_buf	ebuf;
	loff_t          image_size;
	/* mode-specific context fields */
	union {
		struct {
			int		has_block_checksum;
			int		max_block_size;
			struct ss_buf	cbuf;
		};
		struct {
			int		fs_type;
			const char	*interface;
			const char	*devpart;
			const char	*filename;
			loff_t		total_read_len;
			loff_t		file_total_size;
		};
	};
};

enum {
	SSBOOT_LZ4_MEM,
	SSBOOT_ELF_FILE,
	SSBOOT_ELF_MEM,
};

/*
 * LZ4 compression
 *
 * Flags byte:
 * +---------+--------------+----------+---------+----------+----------+---------+
 * |   7-6   |       5      |     4    |    3    |     2    |     1    |    0    |
 * +---------+--------------+----------+---------+----------+----------+---------+
 * | Version |     Block    |   Block  | Content |  Content | Reserved | Dict ID |
 * |         | Independence | Checksum |   Size  | Checksum |          |         |
 * +---------+--------------+----------+---------+----------+----------+---------+
 *
 * Block descriptor byte:
 * +----------+----------+----------+
 * |     7    |   6-5-4  |  3-2-1-0 |
 * +----------+----------+----------+
 * | Reserved |   Block  | Reserved |
 * |          | Max Size |          |
 * +----------+----------+----------+
 */
#define LZ4F_VERSION(x)			(FIELD_GET(GENMASK(7, 6), (x)))
#define LZ4F_BLOCK_INDEP(x)		(FIELD_GET(BIT(5), (x)))
#define LZ4F_HAS_BLOCK_CKSUM(x)		(FIELD_GET(BIT(4), (x)))
#define LZ4F_HAS_CONTENT_SIZE(x)	(FIELD_GET(BIT(3), (x)))
#define LZ4F_MAX_BLOCK_SIZE(x)		(FIELD_GET(GENMASK(6, 4), (x)))

#define LZ4F_FLAGS_RESERVED_MASK	0x03
#define LZ4F_BLOCK_DESC_RESERVED_MASK	0x8F
