/*
 * (C) Copyright 2016 Sony Corporation
 */

#include <boot_time.h>
#include <common.h>
#include <cpu_func.h>
#include <div64.h>
#include <linux/types.h>
#include <fdt_support.h>
#include <fdtdec.h>
#include <asm/global_data.h>

struct boot_time_info {
	phys_addr_t base;
	phys_size_t size;
	bool enabled;
	bool ringbuff;
	u32 max_comment;
	u32 version;
};

static struct boot_time_info bti;

DECLARE_GLOBAL_DATA_PTR;

static struct boot_time_t *pBootTime;
#define BOOT_TIME_ENTRY_SIZE	(sizeof(struct boot_time_entry_t) + bti.max_comment)
#define PENTRY(off)	((struct boot_time_entry_t *)((unsigned long)pBootTime + off))
#define INC_OFFSET(head, off, max)				\
	do {							\
		if (!bti.ringbuff) {				\
			(off) += BOOT_TIME_ENTRY_SIZE;		\
			break;					\
		}						\
		if (((off)+BOOT_TIME_ENTRY_SIZE) >= (max)) {	\
			(off) = (head);				\
		} else {					\
			(off) += BOOT_TIME_ENTRY_SIZE;		\
		}						\
	} while (0)

#define WORD_ALIGNED(addr)	(((unsigned long)(addr) & (4 - 1)) == 0)

static void
boot_time_count_set(struct boot_time_entry_t *pEntry)
{
	unsigned long long t;

	t = (unsigned long long)timer_get_us() * 1000;

	pEntry->count_lo = (u_int32_t)t;
	pEntry->count_hi = (u_int32_t)(t >> 32);
}

void
boot_time_add(char *comment)
{
	struct boot_time_entry_t  *p;
	int nr_entry;

	if (pBootTime == NULL)
		return;

	if (pBootTime->magic != BOOT_TIME_MAGIC) {
		printf("boottime magic is incorrect, entry not added\n");
		return;
	}

	if (!bti.ringbuff && pBootTime->offNext >= pBootTime->offMax)
		return;

	p = PENTRY(pBootTime->offNext);
	boot_time_count_set(p);
	if (comment)
		strncpy(p->comment, comment, bti.max_comment);
	else
		p->comment[0] = '\0';
	INC_OFFSET(pBootTime->offHead, pBootTime->offNext, pBootTime->offMax);
	nr_entry = (pBootTime->offMax - pBootTime->offHead) /
			BOOT_TIME_ENTRY_SIZE;

	if (pBootTime->numWritten < nr_entry)
		pBootTime->numWritten++;
	return;
}

static int fdt_compare_basename(void *blob, int node, char *basename, int len)
{
	char *name, *p;
	int result;

	name = (char *)fdt_get_name(blob, node, &len);
	p = memchr(name, '@', len);
	if (p)
		*p = '\0';
	result = strncmp(name, basename, len);
	if (p)
		*p = '@';

	return result;
}

static int fdt_get_boottime_node(void *blob)
{
	int node, rmem;
	int found = 0;

	rmem = fdt_path_offset(blob, "/reserved-memory");
	if (rmem < 0) {
		printf("Can not open /reserved-memory\n");
		return rmem;
	}

	fdt_for_each_subnode(node, blob, rmem) {
		if (fdt_compare_basename(blob, node, "boottime",
					sizeof("boottime")) == 0) {
			found = 1;
			break;
		}
	}

	if (!found) {
		printf("can not found boottime node\n");
		return -FDT_ERR_NOTFOUND;
	}

	if (!fdtdec_get_is_enabled(blob, node)) {
		printf("boottime is disabled\n");
		return -FDT_ERR_EXISTS;
	}

	return node;
}

#ifdef CONFIG_OF_CONTROL
static int boot_time_info_init(struct boot_time_info *info)
{
	int ret, node;
	void *blob = (void *)gd->fdt_blob;
	struct fdt_resource res;

	node = fdt_get_boottime_node(blob);
	if (node < 0)
		return node;
	info->enabled = true;

	info->version = fdt_getprop_u32_default_node(blob, node, 0,
				"version", 0);
	if (info->version != BOOT_TIME_VERSION) {
		/* U-Boot only support version 2 */
		printf("version %d is not supported\n", info->version);
		return -1;
	}

	ret = fdt_get_resource(blob, node, "reg", 0, &res);
	if (ret < 0) {
		printf("No memory for boottime\n");
		return ret;
	}

	info->base = res.start;
	info->size = fdt_resource_size(&res);

	info->ringbuff = fdtdec_get_bool(blob, node, "ring-buffer");
	info->max_comment = fdt_getprop_u32_default_node(blob, node,
				0, "max-comment", 24);
	if (info->max_comment & 3) {
		printf("max-comment should be 4 bytes aligned value\n");
		return -1;
	}
	if (info->max_comment < 24 || info->max_comment > 128) {
		printf("max-comment should be in 24 ~ 128\n");
		return -1;
	}

	return 0;
}
#else
static int boot_time_info_init(struct boot_time_info *info)
{
	info->enabled = true;
	info->base = CONFIG_DEFAULT_BOOT_TIME_BASE;
	info->size = CONFIG_DEFAULT_BOOT_TIME_SIZE;
#ifdef CONFIG_BOOT_TIME_RINGBUFFER
	info->ringbuff = true;
#endif
	info->max_comment = CONFIG_BOOT_TIME_MAX_COMMENT;
	info->version = BOOT_TIME_VERSION;

	return 0;
}
#endif

static void boot_time_init(void)
{
	u_int32_t header_size;
	int nr_entry;

	memset(pBootTime, 0, bti.size);

	header_size = (unsigned long)&(pBootTime->first_entry) -
			(unsigned long)pBootTime;

	/* force offMax align to entry start address */
	nr_entry = (bti.size - header_size) / BOOT_TIME_ENTRY_SIZE;
	pBootTime->offMax = header_size + nr_entry * BOOT_TIME_ENTRY_SIZE;

	/* initialize boot-time data region */
	pBootTime->magic = BOOT_TIME_MAGIC;
	pBootTime->offHead = header_size;
	pBootTime->offNext = pBootTime->offHead;
	pBootTime->numWritten = 0;
	pBootTime->max_comment = bti.max_comment;

	flush_dcache_all();
}

static void boot_time_clear(void)
{
	if (pBootTime->magic != BOOT_TIME_MAGIC) {
		int c;

		printf("Do you want to force clear? y/n: ");

		c = getchar();
		printf("%c\n", c);

		if (c != 'y')
			return;
	}

	boot_time_init();
}

int initr_boottime(void)
{
	int ret;

	ret = boot_time_info_init(&bti);
	if (ret < 0) {
		printf("fail to init boot_time_info(%d)\n", ret);
		/* If return error, board will be reset */
		return 0;
	};

	pBootTime = (struct boot_time_t *)bti.base;

	if ((pBootTime->magic != BOOT_TIME_MAGIC)
	   || ((u_int32_t)pBootTime->offHead > (u_int32_t)pBootTime->offMax)
	   || ((u_int32_t)pBootTime->offNext > (u_int32_t)pBootTime->offMax)
	   || (!WORD_ALIGNED(pBootTime->offHead))
	   || (!WORD_ALIGNED(pBootTime->offNext))
	   || (pBootTime->max_comment != bti.max_comment)) {
		boot_time_init();
	}

	return 0;
}

int
shell_boottime(int argc, char *const argv[])
{
	struct boot_time_entry_t *p;
	u_int32_t off;
	int loop, num;
	int nr_entry;

	if (!bti.enabled) {
		printf("boottime is disabled in dts\n");
		return 0;
	}

	if (pBootTime->magic != BOOT_TIME_MAGIC) {
		printf("Other function may save its data to boottime region\n");
		printf("Please consider changing boottime region address\n");
	}

	if (argc > 1) {
		if (strcmp(argv[1], "-c") == 0) {
			boot_time_clear();
			return 0;
		}
	}

	if (pBootTime->magic != BOOT_TIME_MAGIC) {
		printf("boottime magic is incorrect\n");
		return 0;
	}

	if (!bti.ringbuff) {
		num = (pBootTime->offNext - pBootTime->offHead) / BOOT_TIME_ENTRY_SIZE;
		off = pBootTime->offHead;
	} else {
		nr_entry = (pBootTime->offMax - pBootTime->offHead) /
			BOOT_TIME_ENTRY_SIZE;
		num = pBootTime->numWritten;
		if (pBootTime->numWritten >= nr_entry)
			off = pBootTime->offNext;
		else
			off = pBootTime->offHead;
	}

	for (loop = 0; loop < num; loop++) {
		int i;
		int space = 0;
		int usec;

		p = PENTRY(off);
		usec = lldiv(((unsigned long long)p->count_hi << 32) |
				p->count_lo, 1000);
		printf("%7u.%03u ms : ", usec / 1000, usec % 1000);
		for (i = 0; i < bti.max_comment; i++) {
			if (space || p->comment[i] == '\0' || p->comment[i] == '\n')
				space = 1;
			else
				putc(p->comment[i]);
		}
		printf("\n");
		INC_OFFSET(pBootTime->offHead, off, pBootTime->offMax);
	}
	printf("\n");
	return 0;
}

static int fdt_create_boottime_node(void *fdt, struct boot_time_info *info)
{
	int ret, node;
	uint32_t phdl;
	struct fdt_memory mem;

	mem.start = info->base;
	mem.end = info->base + info->size - 1;

	ret = fdtdec_add_reserved_memory(fdt, "boottime", &mem, NULL, 0, &phdl,
					 0);
	if (ret < 0) {
		printf("Can't create node in /reserved-memory.\n");
		return ret;
	}

	node = fdt_node_offset_by_phandle(fdt, phdl);
	if (node < 0) {
		printf("Can't open the boottime node\n");
		return node;
	}

	if (fdt_compare_basename(fdt, node, "boottime", sizeof("boottime"))) {
		printf("boottime: %p - %p already reserved by %s\n",
			(void *)mem.start, (void *)mem.end,
			fdt_get_name(fdt, node, NULL));
		return -1;
	}

	return node;
}

int fdt_fixup_boottime(void *fdt)
{
	int ret, node;

	if (!bti.enabled)
		return 0;

	node = fdt_get_boottime_node(fdt);
	if (node == -FDT_ERR_NOTFOUND) {
		node = fdt_create_boottime_node(fdt, &bti);
		if (node < 0)
			return node;
	} else if (node == -FDT_ERR_EXISTS) { /* exist a disabled node */
		/* Disabled in kernel, don't pass the configurations */
		printf("boottime is disabled in kernel\n");
		return 0;
	} else {
		int parent = fdt_parent_offset(fdt, node);

		fdt_delprop(fdt, node, "reg");
		ret = fdt_appendprop_addrrange(fdt, parent, node, "reg",
				bti.base, bti.size);
		if (ret < 0) {
			printf("boottime: fail to update reg in kernel dts\n");
			return ret;
		}
	}

	ret = fdt_setprop_u32(fdt, node, "version", bti.version);
	if (ret < 0) {
		printf("Can't create property: version\n");
		return ret;
	}

	ret = fdt_setprop_u32(fdt, node, "max-comment", bti.max_comment);
	if (ret < 0) {
		printf("Can't create property: max-comment\n");
		return ret;
	}

	if (bti.ringbuff) {
		ret = fdt_setprop(fdt, node, "ring-buffer", NULL, 0);
		if (ret < 0) {
			printf("Can't create property: ring-buffer\n");
			return ret;
		}
	} else
		fdt_delprop(fdt, node, "ring-buffer");

	return 0;
}

#if defined(CONFIG_LMB) && !defined(CONFIG_OF_CONTROL)
#include <lmb.h>

int boot_time_lmb_reserve(struct lmb *lmb)
{
	return lmb_reserve(lmb, bti.base, bti.size);
}
#endif
